# Copyright 2015 Vladimir Rutsky <vladimir@rutsky.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Library meta information.

This module must be stand-alone executable.
"""

__title__ = "aiohttp-cors"
__version__ = "0.7.0"
__author__ = "Vladimir Rutsky and aio-libs team"
__email__ = "vladimir@rutsky.org"
__summary__ = "CORS support for aiohttp"
__uri__ = "https://github.com/aio-libs/aiohttp-cors"
__license__ = "Apache License, Version 2.0"
__copyright__ = "2015-2018 {}".format(__author__)
